//
// Generated by JTB 1.2.2
//
package visitor;

import syntaxtree.*;
import java.util.*;

/**
 * A skeleton output formatter for your language grammar.  Using the
 * add() method along with force(), indent(), and outdent(), you can
 * easily specify how this visitor will format the given syntax tree.
 * See the JTB documentation for more details.
 *
 * Pass your syntax tree to this visitor, and then to the TreeDumper
 * visitor in order to "pretty print" your tree.
 */
public class TreeFormatter extends DepthFirstVisitor {
   private Vector cmdQueue = new Vector();
   private boolean lineWrap;
   private int wrapWidth;
   private int indentAmt;
   private int curLine = 1;
   private int curColumn = 1;
   private int curIndent = 0;

   /**
    * The default constructor assumes an indentation amount of 3 spaces
    * and no line-wrap.  You may alternately use the other constructor to
    * specify your own indentation amount and line width.
    */
   public TreeFormatter() { this(3, 0); }

   /**
    * This constructor accepts an indent amount and a line width which is
    * used to wrap long lines.  If a token's beginColumn value is greater
    * than the specified wrapWidth, it will be moved to the next line and
    * indented one extra level.  To turn off line-wrapping, specify a
    * wrapWidth of 0.
    *
    * @param   indentAmt   Amount of spaces per indentation level.
    * @param   wrapWidth   Wrap lines longer than wrapWidth.  0 for no wrap.
    */
   public TreeFormatter(int indentAmt, int wrapWidth) {
      this.indentAmt = indentAmt;
      this.wrapWidth = wrapWidth;

      if ( wrapWidth > 0 )
         lineWrap = true;
      else
         lineWrap = false;
   }

   /**
    * Accepts a NodeListInterface object and performs an optional format
    * command between each node in the list (but not after the last node).
    */
   protected void processList(NodeListInterface n) {
      processList(n, null);
   }

   protected void processList(NodeListInterface n, FormatCommand cmd) {
      for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
         ((Node)e.nextElement()).accept(this);
         if ( cmd != null && e.hasMoreElements() )
            cmdQueue.addElement(cmd);
      }
   }

   /**
    * A Force command inserts a line break and indents the next line to
    * the current indentation level.  Use "add(force());".
    */
   protected FormatCommand force() { return force(1); }
   protected FormatCommand force(int i) {
      return new FormatCommand(FormatCommand.FORCE, i);
   }

   /**
    * An Indent command increases the indentation level by one (or a
    * user-specified amount).  Use "add(indent());".
    */
   protected FormatCommand indent() { return indent(1); }
   protected FormatCommand indent(int i) {
      return new FormatCommand(FormatCommand.INDENT, i);
   }

   /**
    * An Outdent command is the reverse of the Indent command: it reduces
    * the indentation level.  Use "add(outdent());".
    */
   protected FormatCommand outdent() { return outdent(1); }
   protected FormatCommand outdent(int i) {
      return new FormatCommand(FormatCommand.OUTDENT, i);
   }

   /**
    * A Space command simply adds one or a user-specified number of
    * spaces between tokens.  Use "add(space());".
    */
   protected FormatCommand space() { return space(1); }
   protected FormatCommand space(int i) {
      return new FormatCommand(FormatCommand.SPACE, i);
   }

   /**
    * Use this method to add FormatCommands to the command queue to be
    * executed when the next token in the tree is visited.
    */
   protected void add(FormatCommand cmd) {
      cmdQueue.addElement(cmd);
   }

   /**
    * Executes the commands waiting in the command queue, then inserts the
    * proper location information into the current NodeToken.
    *
    * If there are any special tokens preceding this token, they will be
    * given the current location information.  The token will follow on
    * the next line, at the proper indentation level.  If this is not the
    * behavior you want from special tokens, feel free to modify this
    * method.
    */
   public void visit(NodeToken n) {
      for ( Enumeration e = cmdQueue.elements(); e.hasMoreElements(); ) {
         FormatCommand cmd = (FormatCommand)e.nextElement();
         switch ( cmd.getCommand() ) {
         case FormatCommand.FORCE :
            curLine += cmd.getNumCommands();
            curColumn = curIndent + 1;
            break;
         case FormatCommand.INDENT :
            curIndent += indentAmt * cmd.getNumCommands();
            break;
         case FormatCommand.OUTDENT :
            if ( curIndent >= indentAmt )
               curIndent -= indentAmt * cmd.getNumCommands();
            break;
         case FormatCommand.SPACE :
            curColumn += cmd.getNumCommands();
            break;
         default :
            throw new TreeFormatterException(
               "Invalid value in command queue.");
         }
      }

      cmdQueue.removeAllElements();

      //
      // Handle all special tokens preceding this NodeToken
      //
      if ( n.numSpecials() > 0 )
         for ( Enumeration e = n.specialTokens.elements();
               e.hasMoreElements(); ) {
            NodeToken special = (NodeToken)e.nextElement();

            //
            // -Place the token.
            // -Move cursor to next line after the special token.
            // -Don't update curColumn--want to keep current indent level.
            //
            placeToken(special, curLine, curColumn);
            curLine = special.endLine + 1;
         }

      placeToken(n, curLine, curColumn);
      curLine = n.endLine;
      curColumn = n.endColumn;
   }

   /**
    * Inserts token location (beginLine, beginColumn, endLine, endColumn)
    * information into the NodeToken.  Takes into account line-wrap.
    * Does not update curLine and curColumn.
    */
   private void placeToken(NodeToken n, int line, int column) {
      int length = n.tokenImage.length();

      //
      // Find beginning of token.  Only line-wrap for single-line tokens
      //
      if ( !lineWrap || n.tokenImage.indexOf('\n') != -1 ||
           column + length <= wrapWidth )
         n.beginColumn = column;
      else {
         ++line;
         column = curIndent + indentAmt + 1;
         n.beginColumn = column;
      }

      n.beginLine = line;

      //
      // Find end of token; don't count \n if it's the last character
      //
      for ( int i = 0; i < length; ++i ) {
         if ( n.tokenImage.charAt(i) == '\n' && i < length - 1 ) {
            ++line;
            column = 1;
         }
         else
            ++column;
      }

      n.endLine = line;
      n.endColumn = column;
   }

   //
   // User-generated visitor methods below
   //

   /**
    * f0 -> ( statement() )*
    * f1 -> <END>
    */
   public void visit(label n) {
      if ( n.f0.present() ) {
         processList(n.f0);
      }
      n.f1.accept(this);
   }

   /**
    * f0 -> assignment_stmt()
    *       | pointer_stmt()
    *       | object_stmt()
    *       | group_stmt()
    */
   public void visit(statement n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> attribute_identifier()
    * f1 -> <EQUAL_SIGN>
    * f2 -> value()
    */
   public void visit(assignment_stmt n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * f0 -> <ID>
    */
   public void visit(attribute_identifier n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <LETTER>
    *       | <ID>
    */
   public void visit(identifier n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <BEGIN_OB>
    * f1 -> <EQUAL_SIGN>
    * f2 -> object_identifier()
    * f3 -> ( statement() )*
    * f4 -> <END_OB>
    * f5 -> ( <EQUAL_SIGN> object_identifier() )?
    */
   public void visit(object_stmt n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      if ( n.f3.present() ) {
         processList(n.f3);
      }
      n.f4.accept(this);
      if ( n.f5.present() ) {
         n.f5.accept(this);
      }
   }

   /**
    * f0 -> <ID>
    */
   public void visit(object_identifier n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <BEGIN_GR>
    * f1 -> <EQUAL_SIGN>
    * f2 -> group_identifier()
    * f3 -> ( statement() )*
    * f4 -> <END_GR>
    * f5 -> ( <EQUAL_SIGN> group_identifier() )?
    */
   public void visit(group_stmt n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      if ( n.f3.present() ) {
         processList(n.f3);
      }
      n.f4.accept(this);
      if ( n.f5.present() ) {
         n.f5.accept(this);
      }
   }

   /**
    * f0 -> <ID>
    */
   public void visit(group_identifier n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <POINTER>
    * f1 -> object_identifier()
    * f2 -> <EQUAL_SIGN>
    * f3 -> value()
    */
   public void visit(pointer_stmt n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   /**
    * f0 -> sequence_value()
    *       | scalar_value()
    *       | set_value()
    */
   public void visit(value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> numeric_value()
    *       | symbolic_value()
    *       | text_string_value()
    *       | date_time_value()
    */
   public void visit(scalar_value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> numeric_option()
    *       | numeric_option() units_expression()
    */
   public void visit(numeric_value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> integer()
    *       | based_integer()
    *       | <FLOATING_POINT_LITERAL>
    */
   public void visit(numeric_option n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> units_factor()
    * f1 -> ( mult_op() units_factor() )*
    */
   public void visit(units_expression n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
   }

   /**
    * f0 -> <MULTIPLICATION_DIVISION_SIGN>
    */
   public void visit(mult_op n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> units_identifier()
    *       | units_identifier() exp_op() integer()
    */
   public void visit(units_factor n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <ID>
    */
   public void visit(units_identifier n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <EXPONENT_SIGN>
    */
   public void visit(exp_op n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> time()
    *       | date() ( <TIME_CONSTANT> time() )*
    *       | <DATE_TIME>
    */
   public void visit(date_time_value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> quoted_text()
    */
   public void visit(text_string_value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> identifier()
    *       | quoted_symbol()
    */
   public void visit(symbolic_value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> sequence_1D()
    *       | sequence_2D()
    */
   public void visit(sequence_value n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <LEFT_PARAN>
    * f1 -> scalar_value()
    * f2 -> ( <COMMA> scalar_value() )*
    * f3 -> <RIGHT_PARAN>
    */
   public void visit(sequence_1D n) {
      n.f0.accept(this);
      n.f1.accept(this);
      if ( n.f2.present() ) {
         processList(n.f2);
      }
      n.f3.accept(this);
   }

   /**
    * f0 -> <LEFT_PARAN>
    * f1 -> ( sequence_1D() )+
    * f2 -> <RIGHT_PARAN>
    */
   public void visit(sequence_2D n) {
      n.f0.accept(this);
      processList(n.f1);
      n.f2.accept(this);
   }

   /**
    * f0 -> <LEFT_BRACE>
    * f1 -> scalar_value()
    * f2 -> ( <COMMA> scalar_value() )*
    * f3 -> <RIGHT_BRACE>
    */
   public void visit(set_value n) {
      n.f0.accept(this);
      n.f1.accept(this);
      if ( n.f2.present() ) {
         processList(n.f2);
      }
      n.f3.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    *       | <SIGNED_INTEGER>
    */
   public void visit(integer n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <PLUS_MINUS_SIGN>
    */
   public void visit(sign n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> ( <DIGIT> )+
    */
   public void visit(unsigned_integer n) {
      processList(n.f0);
   }

   /**
    * f0 -> <DIGIT>
    */
   public void visit(digit n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> radix() <HASH_SIGN> ( extended_digit() )+ <HASH_SIGN>
    *       | radix() <HASH_SIGN> sign() ( extended_digit() )+ <HASH_SIGN>
    */
   public void visit(based_integer n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> digit()
    *       | letter()
    */
   public void visit(extended_digit n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <LETTER>
    */
   public void visit(letter n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    */
   public void visit(radix n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> ( sign() )? unscaled_real()
    *       | ( sign() )? scaled_real()
    */
   public void visit(real n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer() <DOT_SIGN> unsigned_integer()
    *       | unsigned_integer() <DOT_SIGN>
    *       | <DOT_SIGN> unsigned_integer()
    */
   public void visit(unscaled_real n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unscaled_real()
    * f1 -> exponent()
    */
   public void visit(scaled_real n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * f0 -> <EXP_E_C> integer()
    *       | <EXP_E_L> integer()
    */
   public void visit(exponent n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> year_day()
    */
   public void visit(date n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <YEAR_DAY>
    */
   public void visit(year_day n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> year()
    * f1 -> <MINUS_SIGN>
    * f2 -> month()
    * f3 -> <MINUS_SIGN>
    * f4 -> day()
    */
   public void visit(year_month_day n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    */
   public void visit(year n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    */
   public void visit(month n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    */
   public void visit(day n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> hour_min_sec() ( <Z_SIGN> )?
    *       | zoned_time()
    */
   public void visit(time n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> hour_min_sec()
    */
   public void visit(local_time n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> hour_min_sec()
    * f1 -> <Z_SIGN>
    */
   public void visit(utc_time n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * f0 -> hour_min_sec()
    * f1 -> zone_offset()
    */
   public void visit(zoned_time n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   /**
    * f0 -> <HOUR_MIN>
    */
   public void visit(hour_min_sec n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    */
   public void visit(hour n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    */
   public void visit(minute n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> unsigned_integer()
    *       | unscaled_real()
    */
   public void visit(second n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> sign()
    * f1 -> hour()
    * f2 -> ( <COLON_SIGN> minute() )?
    */
   public void visit(zone_offset n) {
      n.f0.accept(this);
      n.f1.accept(this);
      if ( n.f2.present() ) {
         n.f2.accept(this);
      }
   }

   /**
    * f0 -> date()
    * f1 -> <TIME_CONSTANT>
    * f2 -> time()
    */
   public void visit(date_time n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * f0 -> <QUOTED_STRING>
    */
   public void visit(quoted_text n) {
      n.f0.accept(this);
   }

   /**
    * f0 -> <SQUOTE_SIGN>
    * f1 -> ( character() )*
    * f2 -> <SQUOTE_SIGN>
    */
   public void visit(quoted_symbol n) {
      n.f0.accept(this);
      if ( n.f1.present() ) {
         processList(n.f1);
      }
      n.f2.accept(this);
   }

   /**
    * f0 -> <LETTER>
    *       | <DIGIT>
    *       | <SPECIAL_CHARACTER>
    *       | <SPACING_CHARACTER>
    *       | <FORMAT_EFFECTOR>
    *       | <OTHER_CHARACTER>
    */
   public void visit(character n) {
      n.f0.accept(this);
   }

}

class FormatCommand {
   public static final int FORCE = 0;
   public static final int INDENT = 1;
   public static final int OUTDENT = 2;
   public static final int SPACE = 3;

   private int command;
   private int numCommands;

   FormatCommand(int command, int numCommands) {
      this.command = command;
      this.numCommands = numCommands;
   }

   public int getCommand()             { return command; }
   public int getNumCommands()         { return numCommands; }
   public void setCommand(int i)       { command = i; }
   public void setNumCommands(int i)   { numCommands = i; }
}

class TreeFormatterException extends RuntimeException {
   TreeFormatterException()         { super(); }
   TreeFormatterException(String s) { super(s); }
}
